#include <Wire.h>
#include <Adafruit_MCP23017.h>
#include <Servo.h> 

// Comment out for no debug.
#define DEBUG 1

// Array to hold the state transition table.
char stt[5][9] = {{'A', 'A', 'A', 'B', 'B', 'B', 'C', 'C', 'C'},
                  {'1', '0', 'b', '1', '0', 'b', '1', '1', 'b'},
                  {' ', ' ', 'b', ' ', ' ', 'b', ' ', ' ', 'b'},
                  {' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '},
                  {' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '}};

// Array to keep track of the tape values.
char tape[] = {'b', '0', '0', '0', '0', '0', '0', '0', '0', 'b'};

// Read/write head position.
int head = 1; 

// Define the transition steps.
#define STATE 0
#define READ  1
#define WRITE 2
#define MOVE  3
#define GOTO  4

// Set true when the PLAY button has been pressed.
boolean nextStep = true;

// Enable RUN mode.
boolean runMode = false;

// Speed when in RUN mode;
int runSpeed = 200;

// Flag to indicate the the system has just started up or the RESET button has been pressed.
boolean reset = true;

// The current step in the transition.
int currentStep = STATE;

// The current transition we are working on.
int transition;

// Current state (starting with the default state A).
int state = 'A';

// Define the transitions.
#define tA1 0
#define tA0 1
#define tAb 2
#define tB1 3
#define tB0 4
#define tBb 5
#define tC1 6
#define tC0 7
#define tCb 8

// Instantiate a servo object for each connected servo. 
Servo FLIP_BIT_1_SERVO;    
Servo FLIP_BIT_2_SERVO;
Servo FLIP_BIT_3_SERVO;
Servo FLIP_BIT_4_SERVO;
Servo FLIP_BIT_5_SERVO;
Servo FLIP_BIT_6_SERVO;
Servo FLIP_BIT_7_SERVO;
Servo FLIP_BIT_8_SERVO;   

// Servo angles used to display a zero or a one.
int ZERO = 0;
int ONE = 180;

// Outputs that control the servos.
int FLIP_BIT_1_OUTPUT = 2;
int FLIP_BIT_2_OUTPUT = 3;
int FLIP_BIT_3_OUTPUT = 4;
int FLIP_BIT_4_OUTPUT = 5;
int FLIP_BIT_5_OUTPUT = 6;
int FLIP_BIT_6_OUTPUT = 7;
int FLIP_BIT_7_OUTPUT = 8;
int FLIP_BIT_8_OUTPUT = 9;

// Hall Effect sensor inputs.
int wA1 = 22;  // WRITE.
int wA0 = 23;
int wB1 = 24;
int wB0 = 25;
int wC1 = 26;
int wC0 = 27;

int mA1 = 28;  // MOVE.
int mA0 = 29;
int mAb = 30;
int mB1 = 31;
int mB0 = 32;
int mBb = 33;
int mC1 = 34;
int mC0 = 35;
int mCb = 36;

int g1A1 = 37; // GOTO1.
int g1A0 = 38;
int g1Ab = 39;
int g1B1 = 40;
int g1B0 = 41;
int g1Bb = 42;
int g1C1 = 43;
int g1C0 = 44;
int g1Cb = 45;

int g2A1 = 46; // GOTO1.
int g2A0 = 47;
int g2Ab = 48;
int g2B1 = 49;
int g2B0 = 50;
int g2Bb = 51;
int g2C1 = 52;
int g2C0 = 53;
int g2Cb = 10; // Don't forget about this exception!

// Finate State Machine status lights.
int lC = 15;
int lB = 14;
int lA = 13;
int lREAD = 12;
int lWRITE = 11;
int lMOVE = 10;
int lGOTO = 9;
int lA1 = 8;
int lA0 = 0;
int lAb = 1;
int lB1 = 2;
int lB0 = 3;
int lBb = 4;
int lC1 = 5;
int lC0 = 6;
int lCb = 7;

// Finate State Machine Panel other controls.
int lHALT = 11;
int bPLAY = 12;
int bRUN_STEP = 13;

// Play Panel buttons.
int bRight = 15;
int bFlip = 14;
int bLeft = 13;
int bReset = 12;

// Current tape head position and mapping to output lamps.
int headMapper[] = {9, 8, 0, 1, 2, 3, 4, 5, 6, 7};

// Object to manage MCP23017 I/O Extender for the Tape Panel.
Adafruit_MCP23017 mcp;

// Object to manage MCP23017 I/O Extender for the Finite State Machine Panel.
Adafruit_MCP23017 mcp2;

void setup() {
#ifdef DEBUG
  // Initialize serial debugging.
  Serial.begin(115000);
  Serial.println("Turing Machine Demonstrator...");
#endif
  // Attach the servos. Note the min and max pulse widths might have to be tweaked
  // to get the bits level.
  FLIP_BIT_1_SERVO.attach(FLIP_BIT_1_OUTPUT,565,2547); 
  FLIP_BIT_1_SERVO.write(ZERO);
  delay(50);
  FLIP_BIT_2_SERVO.attach(FLIP_BIT_2_OUTPUT,560,2500);
  FLIP_BIT_2_SERVO.write(ZERO);
  delay(50);
  FLIP_BIT_3_SERVO.attach(FLIP_BIT_3_OUTPUT,560,2500);
  FLIP_BIT_3_SERVO.write(ZERO);
  delay(50);
  FLIP_BIT_4_SERVO.attach(FLIP_BIT_4_OUTPUT,560,2500);
  FLIP_BIT_4_SERVO.write(ZERO);
  delay(50);
  FLIP_BIT_5_SERVO.attach(FLIP_BIT_5_OUTPUT,560,2500);
  FLIP_BIT_5_SERVO.write(ZERO);
  delay(50);
  FLIP_BIT_6_SERVO.attach(FLIP_BIT_6_OUTPUT,555,2500);
  FLIP_BIT_6_SERVO.write(ZERO);
  delay(50);
  FLIP_BIT_7_SERVO.attach(FLIP_BIT_7_OUTPUT,560,2460);
  FLIP_BIT_7_SERVO.write(ZERO); 
  delay(50);
  FLIP_BIT_8_SERVO.attach(FLIP_BIT_8_OUTPUT,560,2500);
  FLIP_BIT_8_SERVO.write(ZERO);  
  delay(50);
  
  // Initialize the I2C interfaces.
  mcp.begin();
  mcp2.begin(1);

  // Setup all the output pins on the Tape Panel.
  for (int i = 0; i < 10; i++) {
    mcp.pinMode(i, OUTPUT);;
  }

  // Setup all the output pins on the Finite State Machine Panel.
  for (int i = 0; i < 16; i++) {
    mcp2.pinMode(i, OUTPUT);;
  }

  // Other control pins on the Finite State Machine Panel.
  pinMode(lHALT, OUTPUT);
  pinMode(bPLAY, INPUT_PULLUP);
  pinMode(bRUN_STEP, INPUT_PULLUP);
  
  // Clean start.
  resetHead();
  resetTape();
  clearTransitionLamps();
  
  // Setup all the Tape Panel input pins as pulled up high.
  mcp.pinMode(bReset, INPUT);
  mcp.pullUp(bReset, HIGH);
  mcp.pinMode(bLeft, INPUT);
  mcp.pullUp(bLeft, HIGH);
  mcp.pinMode(bFlip, INPUT);
  mcp.pullUp(bFlip, HIGH);
  mcp.pinMode(bRight, INPUT);
  mcp.pullUp(bRight, HIGH);

  // Setup all the Hall Effect sensors as pulled up high inputs.
  for (int i = wA1; i <= g2C0; i++) {
     pinMode(i, INPUT_PULLUP);
  }
  pinMode(g2Cb, INPUT_PULLUP);

  // Load the transition table based on the Finite State Machine inputs..
  loadTransitionTable();

  // Set the servos to their zero start position.
  //resetServos();
}

void loop(){

  // Check the Tape Panel inputs.
  int newHead = head;
  if (mcp.digitalRead(bRight) == 0) {
    if (!reset) {
      runSpeed -= 10;
      if (runSpeed < 50) runSpeed = 50;
    } else {
      // Mode head left.
      if (newHead > 0) {
        newHead--;
        showHead(newHead);
        debounce(bRight);
      }
    }
  } else if (mcp.digitalRead(bLeft) == 0) {
    if (!reset) {
      runSpeed += 10;
      if (runSpeed > 500) runSpeed = 500;
    } else {
      // Move head right.
      if (newHead < 9) {
        newHead++;
        showHead(newHead);
        debounce(bLeft);
      }
    }
  } else if (mcp.digitalRead(bFlip) == 0) {
    if (reset) {
      // Flip bit.
      flipCurrentBit(head);
      debounce(bFlip);
    }
  } else if (mcp.digitalRead(bReset) == 0) {
      // Reset the tape head and flip bits.
      resetServos();
      resetHead();
      resetTape();
      state = 'A';
      clearTransitionLamps();
      digitalWrite(lHALT, 0);
      currentStep = STATE;
      nextStep = true;
      reset = true;
      loadTransitionTable();
      debounce(bReset);
  }

  // Check the Finite State Machine Panel inputs.
  if (digitalRead(bPLAY) == 0) {
    if (reset || nextStep == false) {
      // Starting or restarting the FSM.
      nextStep = true;
    } else {
      // Stop the machine if it's already running. 
      nextStep = false;
    }
    reset = false;
    debounce2(bPLAY);
  }
  if (digitalRead(bRUN_STEP) == 0) {
    runMode = true;
  } else {
    runMode = false;
  }
  

  // Check for the HALT state.
  if (state == 'H') {
    digitalWrite(lHALT, 1);
    // Don't let the user continue.
    nextStep = false;
  }

  // The user must press the run button to continue.
  if (nextStep == false || (reset == true && currentStep != STATE)) {
    delay(100);
    return;
  }
  
  // Run the State Machine.
  char cellValue;
  switch(currentStep) {
    case STATE: 
      // Step 1: Set the starting transition based on the current state. Turn on the appropriate state lamp.
      switch(state) {
        case 'A': 
          transition = 0; 
          blinkON(mcp2, lA);
          mcp2.digitalWrite(lB, 0);
          mcp2.digitalWrite(lC, 0);
        break;
        case 'B': 
          transition = 3; 
          mcp2.digitalWrite(lA, 0);
          blinkON(mcp2, lB);
          mcp2.digitalWrite(lC, 0);
        break;
        case 'C': 
          transition = 6; 
          mcp2.digitalWrite(lA, 0);
          mcp2.digitalWrite(lB, 0);
          blinkON(mcp2, lC);
          break;
      }
      blinkON(mcp2, lREAD);
      
      currentStep++;
      if (!runMode) nextStep = false;
    break;
    
    case READ:   
      // Step 2: READ the cell at the current head position.
      cellValue = tape[head];
    
      // Adjust the transition based on the value found.
      switch(cellValue) {
        case '1': break;
        case '0': 
          transition += 1; 
          break;
        case 'b': 
          transition += 2; 
          break;
      }
      mcp2.digitalWrite(lREAD, 0);
      if (transition == 0) {
        blinkON(mcp2, lA1);
      } else {
        blinkON(mcp2, lA0 + transition - 1);
      }
      blinkON(mcp2, lWRITE);
      currentStep++;
      if (!runMode) nextStep = false;
    break;

    case WRITE:
      // Step 3: WRITE the transition write value back to the tape.   Flip the Tape Panel bit if necessary.
      setBit(head, stt[WRITE][transition]);
      mcp2.digitalWrite(lWRITE, 0);
      blinkON(mcp2, lMOVE);
      currentStep++;
      if (!runMode) nextStep = false;
    break;

    case MOVE:
      // Step 4: MOVE the tape head based on the transition move value.
      switch(stt[MOVE][transition]) {
        case 'R': 
          if (head == 0) {
            state = 'H';
          } else {
            showHead(head - 1);
          }
        break; 
        case 'L': 
          if (head == 9) {
            state = 'H';
          } else {
            showHead(head + 1);
          }
        break; 
      }
      mcp2.digitalWrite(lMOVE, 0);
      blinkON(mcp2, lGOTO);
      currentStep++;
      if (!runMode) nextStep = false;
    break;

    case GOTO: 
      // Step 5: GOTO the next state based on the transition value.
      if (state != 'H') {
        state = stt[GOTO][transition];
        currentStep = STATE;
        clearTransitionLamps();
      } 
    break;
  }
  delay(runSpeed);
}

void debounce(int buttonId) {
  // Wait for button up.
  while (mcp.digitalRead(buttonId) == 0) {
    delay(20);
  }
}

void debounce2(int buttonId) {
  // Wait for button up.
  while (digitalRead(buttonId) == 0) {
    delay(20);
  }
}

void showHead(int newHead) {
  // Clear the current head position.
  mcp.digitalWrite(headMapper[head], 0);
  
  // Set the new head position.
  blinkON(mcp, headMapper[newHead]);
  
  // Update the current head position.
  head = newHead;  
}

void resetTape() {
  tape[0] = 'b';
  for (int i = 1; i < 9; i++) {
    tape[i] = '0';
  }
  tape[9] = 'b';
}

void resetHead() {
  mcp.digitalWrite(0, 0);
  mcp.digitalWrite(1, 0);
  mcp.digitalWrite(2, 0);
  mcp.digitalWrite(3, 0);
  mcp.digitalWrite(4, 0);
  mcp.digitalWrite(5, 0);
  mcp.digitalWrite(6, 0);
  mcp.digitalWrite(7, 0);
  mcp.digitalWrite(8, 0);
  mcp.digitalWrite(9, 0);

  // Set first position and show.
  head = 1;
  showHead(head);
}

void resetServos() {
  FLIP_BIT_1_SERVO.write(ZERO);
  delay(50);
  FLIP_BIT_2_SERVO.write(ZERO);
  delay(50);
  FLIP_BIT_3_SERVO.write(ZERO);
  delay(50);
  FLIP_BIT_4_SERVO.write(ZERO);
  delay(50);
  FLIP_BIT_5_SERVO.write(ZERO);
  delay(50);
  FLIP_BIT_6_SERVO.write(ZERO);
  delay(50);
  FLIP_BIT_7_SERVO.write(ZERO);
  delay(50);
  FLIP_BIT_8_SERVO.write(ZERO);
  delay(50);
}

void flipCurrentBit(int head) {
  if (head == 0 || head == 9){
    // Not a flippable bit.
  } else {
    char currentBit = tape[head];
    switch (head-1) {
      case 0:
        if (currentBit == '0') {
          FLIP_BIT_8_SERVO.write(ONE);
        } else {
          FLIP_BIT_8_SERVO.write(ZERO);
        }
      break;
      case 1:
        if (currentBit == '0') {
          FLIP_BIT_7_SERVO.write(ONE);
        } else {
          FLIP_BIT_7_SERVO.write(ZERO);
        }
      break;
      case 2:
        if (currentBit == '0') {
          FLIP_BIT_6_SERVO.write(ONE);
        } else {
          FLIP_BIT_6_SERVO.write(ZERO);
        }
      break;
      case 3:
        if (currentBit == '0') {
          FLIP_BIT_5_SERVO.write(ONE);
        } else {
          FLIP_BIT_5_SERVO.write(ZERO);
        }
      break;
      case 4:
        if (currentBit == '0') {
          FLIP_BIT_4_SERVO.write(ONE);
        } else {
          FLIP_BIT_4_SERVO.write(ZERO);
        }
      break;
      case 5:
        if (currentBit == '0') {
          FLIP_BIT_3_SERVO.write(ONE);
        } else {
          FLIP_BIT_3_SERVO.write(ZERO);
        }
      break;
      case 6:
        if (currentBit == '0') {
          FLIP_BIT_2_SERVO.write(ONE);
        } else {
          FLIP_BIT_2_SERVO.write(ZERO);
        }
      break;
      case 7:
        if (currentBit == '0') {
          FLIP_BIT_1_SERVO.write(ONE);
        } else {
          FLIP_BIT_1_SERVO.write(ZERO);
        }
      break;
    }
    // Remember the new bit position.
    tape[head] = (currentBit=='0')?'1':'0';
  }
  dumpTape();
}

void dumpTape() {
#ifdef DEBUG
  for (int i = 9; i >= 0; i--) {
    Serial.print(tape[i]);
    Serial.print(' ');
  }
  Serial.println();
#endif
}

void setBit(int head, char newBit) {
    char currentBit = tape[head];
#ifdef DEBUG
    Serial.print(head);
    Serial.print(' ');
    Serial.print(newBit);
    Serial.print(' ');
    Serial.println(currentBit);
#endif
    if (newBit != currentBit) {
      switch (head - 1) {
        case 0:
          if (newBit != '0') {
            FLIP_BIT_8_SERVO.write(ONE);
          } else {
            FLIP_BIT_8_SERVO.write(ZERO);
          }
        break;
        case 1:
          if (newBit != '0') {
            FLIP_BIT_7_SERVO.write(ONE);
          } else {
            FLIP_BIT_7_SERVO.write(ZERO);
          }
        break;
        case 2:
          if (newBit != '0') {
            FLIP_BIT_6_SERVO.write(ONE);
          } else {
            FLIP_BIT_6_SERVO.write(ZERO);
          }
        break;
        case 3:
          if (newBit != '0') {
            FLIP_BIT_5_SERVO.write(ONE);
          } else {
            FLIP_BIT_5_SERVO.write(ZERO);
          }
        break;
        case 4:
          if (newBit != '0') {
            FLIP_BIT_4_SERVO.write(ONE);
          } else {
            FLIP_BIT_4_SERVO.write(ZERO);
          }
        break;
        case 5:
          if (newBit != '0') {
            FLIP_BIT_3_SERVO.write(ONE);
          } else {
            FLIP_BIT_3_SERVO.write(ZERO);
          }
        break;
        case 6:
          if (newBit != '0') {
            FLIP_BIT_2_SERVO.write(ONE);
          } else {
            FLIP_BIT_2_SERVO.write(ZERO);
          }
        break;
        case 7:
          if (newBit != '0') {
            FLIP_BIT_1_SERVO.write(ONE);
          } else {
            FLIP_BIT_1_SERVO.write(ZERO);
          }
        break;
      }
      // Remember the new bit position.
      tape[head] = newBit;
   }
}

void loadTransitionTable() {
    // Load the WRITE values.
    setWrite(digitalRead(wA1), tA1);
    setWrite(digitalRead(wA0), tA0);
    setWrite(digitalRead(wB1), tB1);
    setWrite(digitalRead(wB0), tB0);
    setWrite(digitalRead(wC1), tC1);
    setWrite(digitalRead(wC0), tC0);

    // Load the MOVE values.
    setMove(digitalRead(mA1), tA1);
    setMove(digitalRead(mA0), tA0);
    setMove(digitalRead(mAb), tAb);
    setMove(digitalRead(mB1), tB1);
    setMove(digitalRead(mB0), tB0);
    setMove(digitalRead(mBb), tBb);
    setMove(digitalRead(mC1), tC1);
    setMove(digitalRead(mC0), tC0);
    setMove(digitalRead(mCb), tCb);
    

    // Load the GOTO values.
    setGoto(digitalRead(g1A1), digitalRead(g2A1), tA1);
    setGoto(digitalRead(g1A0), digitalRead(g2A0), tA0);
    setGoto(digitalRead(g1Ab), digitalRead(g2Ab), tAb);
    setGoto(digitalRead(g1B1), digitalRead(g2B1), tB1);
    setGoto(digitalRead(g1B0), digitalRead(g2B0), tB0);
    setGoto(digitalRead(g1Bb), digitalRead(g2Bb), tBb);
    setGoto(digitalRead(g1C1), digitalRead(g2C1), tC1);
    setGoto(digitalRead(g1C0), digitalRead(g2C0), tC0);
    setGoto(digitalRead(g1Cb), digitalRead(g2Cb), tCb);
    

#ifdef DEBUG
    // Dump the table.
    for(int i =0; i < 5; i++) {
      for (int j = 0; j < 9; j++) {
        Serial.print(stt[i][j]);
        Serial.print(' ');
      }
      Serial.println();
    }
    Serial.println();
    Serial.println();
#endif
}

void setWrite(int value, int transaction) {
  if (value == 0) {
      stt[WRITE][transaction] = '1';
    } else {
      stt[WRITE][transaction] = '0';
    }
}

void setMove(int value, int transaction) {
  if (value == 0) {
      stt[MOVE][transaction] = 'R';
    } else {
      stt[MOVE][transaction] = 'L';
    }
}

void setGoto(int top, int bottom, int transaction) {
  if (top == 0 && bottom == 0) {
    stt[GOTO][transaction] = 'H';
  } else if (top == 0 && bottom == 1) {
    stt[GOTO][transaction] = 'B';
  } else if (top == 1 && bottom == 0) {
    stt[GOTO][transaction] = 'C';
  } else {
    stt[GOTO][transaction] = 'A';
  }
}

void clearTransitionLamps() {
  if (state == 'H') {
    return;
  }
  mcp2.digitalWrite(lA1, 0);
  mcp2.digitalWrite(lA0, 0);
  mcp2.digitalWrite(lAb, 0);
  mcp2.digitalWrite(lB1, 0);
  mcp2.digitalWrite(lB0, 0);
  mcp2.digitalWrite(lBb, 0);
  mcp2.digitalWrite(lC1, 0);
  mcp2.digitalWrite(lC0, 0);
  mcp2.digitalWrite(lCb, 0);
  mcp2.digitalWrite(lREAD, 0);
  mcp2.digitalWrite(lWRITE, 0);
  mcp2.digitalWrite(lMOVE, 0);
  mcp2.digitalWrite(lGOTO, 0);
}

/**
 * Blink the lamp a couple of times before leaving it ON. Skip the 
 * blinking if the play button is pressd.
 */
void blinkON(Adafruit_MCP23017 mcpX, int pin) {
  mcpX.digitalWrite(pin, 1);
  if (!runMode) {
    for (int i = 0; i < 30; i++) {
      if (digitalRead(bPLAY) == 0) {
        return;
      }
      delay(10);
    }
    mcpX.digitalWrite(pin, 0);
    for (int i = 0; i < 30; i++) {
      if (digitalRead(bPLAY) == 0) {
        mcpX.digitalWrite(pin, 1);
        return;
      }
      delay(10);
    }
    mcpX.digitalWrite(pin, 1);
    for (int i = 0; i < 30; i++) {
      if (digitalRead(bPLAY) == 0) {
        return;
      }
      delay(10);
    }
  }
}
